package TextData::Calc;

#textdata version 1.9.0 (19.8.2008)
#see POD below!

use Math::Trig;
require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(LinearValue ParsedFormula FormulaFunction);

#a hack for Gauss() function to use in formulae
#not fully verified yet  - and not normalized!
our $cache = undef;
sub Gauss
{
	#Polar-Methode
	my $x;
	if(defined $cache)
	{
		$x = $cache;
		undef $cache;
	}
	else
	{
		my ($u1,$u2,$v);
		do
		{
			$u1 = rand();
			$u2 = rand();
			$v = (2*$u1-1)**2+(2*$u2-1)**2;
		}
		while($v >= 1);
		$cache = (2*$u2-1)*sqrt(-2*log($v)/$v);
		$x = (2*$u1-1)*sqrt(-2*log($v)/$v);
	}
	return $x;
}


#LinearValue(x,[x1,x2],[y1,y2])
sub LinearValue
{
	my ($x,$ox,$oy) = @_;
	return $ox->[0] != $ox->[1] ? ( $oy->[0] + ($oy->[1]-$oy->[0])*($x-$ox->[0])/($ox->[1]-$ox->[0]) ) : ( $x == $ox->[0] ? ($oy->[0]+$oy->[1])/2 : undef );
}

#ParsedFormula(text, dataarrayname, pararrayname1, pararrayname2)
#A -> pararray1
#C -> pararray2
#[a,b] -> data[a][b]

sub ParsedFormula
{
	my ($form, $data, $par1, $par2) = @_;
	my @formlines = split("\n", $form);
	my $nnf = '';

	foreach my $formula (@formlines)
	{
		my $nf = '';
		#$ord$ord is not translated correctly but is not correct syntax anyway
		{ # Parse shortcut vars.
			my %defs =
			(
				 'x',   '[0,1]'
				,'y',   '[0,2]'
				,'ord', '[0,1]'
				,'a',   '[0,2]'
				,'b',   '[1,3]'
			);
			#print STDERR "shortcut parsing: $formula\n";
			while($formula =~ /^(.*)([^a-zA-Z\$]|^)(\$?([xab]|ord))([^(a-zA-Z]|$)(.*)$/m)
			{
				#print STDERR "found $4 (def=$defs{$4})\n";
				$formula = $1.$2.$defs{$4}.$5.$6;
			}
			#print STDERR "done shortcut parsing: $formula\n";
		}
		# Match any relevant [...] and stuff before it; parse and cut from formula.
		#print STDERR "formula: $formula\n";
		while($formula =~ s/\A([^[]*[^[a-zA-Z]|)\[\s*(([^[\],]+)(\s*,\s*([^[\],])|)\s*)\s*\]//)
		{
			#print STDERR "results: $1 : $2 : $3 : $4 : $5\n";
			#print STDERR "formula: $formula\n";
			$nf .= $1;
			my $num1 = $3;
			my $num2 = $5;
			unless(defined $num2)
			{
				$num2 = $num1;
				$num1 = 0;
			}
			if(($num1 =~ /^\d+$/) and ($num1 < 0))
			{
				print STDERR "File index $num1 < 0 !\n";
				return undef;
			}
			if($num2 =~ /^\d+$/)
			{
				--$num2;
				if($num2 < 0)
				{
					print STDERR "Dataset index $num2 < 0 !\n";
					return undef;
				}
			}
			else{ $num2 = "($num2)-1"; }
			$nf .= '$'.$data."[$num1][$num2]";		
			#print STDERR "nf:      $nf\n";
		}

		$nf .= $formula;
		$nf =~ s/(^|[^\$a-zA-Z])A(\d+)/$1\$$par1\[$2\]/g;	
		$nf =~ s/(^|[^\$a-zA-Z])C(\d+)/$1\$$par2\[$2\]/g;
		if($nnf ne ''){ $nnf .= "\n"; }
		$nnf .= $nf;
	}
	return $nnf;
}

#(formula, be_verbose)
sub FormulaFunction
{
	my ($formula,$verb) = @_;
	my $pf = ParsedFormula($formula, 'fd->', 'A->', 'C->');
	unless(defined $pf)
	{
		$@ = "TextData::Calc: Error parsing the formula!";
		return undef;
	}
	print STDERR "Formula code: ",$pf,"\n" if $verb;
	return eval 'sub { my ($fd, $A, $C) = @_; '.$pf.' ; return 0; }';
}

1;
__END__

=head1 NAME

TextData::Calc -  helper package for some calculations

=head1 SYNOPSIS

	use TextData::Calc qw(LinearValue FormulaFunction);
		
	#inter- or extrapolation with known data points ($x1,$y1) and ($x2,$y2)
	$y = LinearValue($x,[$x1,$x2],[$y1,$y2]);
	my $ff = FormulaFunction($textualformula);
	my $value = &$ff(\@datasets);

=head1 DESCRIPTION

This is a little library for TextData; it contains routines that in fact are too general to strictly belong to TextData... but here they are.

=head2 Functions

=over * 4

=item * LinearValue($x,[$x1,$x2],[$y1,$y2]) -> $y

does simple linear inter-/extrapolation of the $y for a given $x based on two known points. It is needed by the TextData::File class for, well, interpolation.

=item * FormulaFunction($formula) -> \&function

This takes some formula as text and constructs a function that evaluates this formula. The created function takes three arrays (references) as arguments: A two-dimensional array and two one-dimensional arrays with some data needed for calculation.
This facility is created for and used by the program named txdcalc to provide flexible data manipulation. So, see that one for formula syntax and semantics.

=item * ParsedFormula($formula, $dataname, $par1name, $par2name) -> $perl_code

is the parser for the $formula that just replaces the different short forms for data and constants with proper PERL code using the last three parameters for the array names. It's used internally by FormulaFunction. If you really want to use it explicitly, then it shouldn't be hard to figure the usage out by looking at the source of FormulaFunction.

=back

=head1 LICENSE

This package is open source software under the terms of the Artistic License. Free use respecting the author;-)

=head1 AUTHOR

Thomas Orgis <thomas@orgis.org>

=head1 WEB

See	http://thomas.orgis.org/scripterei for updates and lots of other (un)related stuff.
