package TextData::Lines;

#note: linecheck should also be able to run in non-strict mode (???)
#hint: lineData in non-strict mode could honor the set separator.
#change: fixed bug with quoted title line (last quote was included where it should not)

#textdata version 1.9.0 (19.8.2008)
#see POD below!

my $trenner = '\s+|,|;';
my $ntrenner = '[^\s,;]+';
my $lend = '[\012\015]+';
my $nlend = '[^\012\015]';

my %defaults = 
(
	'separator',"\t",
	'lineend', "\n",
	'comment', '#',
	'numregex', '\+?\-?\d*\.?\d+e?E?\+?\-?\d*',
	'numberformat',[],
	'comregex','[#%]*[^\S\012\015]*',
	'quote',1
);

sub new
{
	my $class = shift;
	my $self = {};
	bless $self, $class;
	$self->{gotconfig} = shift;
	foreach my $n (@{$self->{gotconfig}->{numberformat}})
	{
		$n =~ s/^([^\%])/\%$1/;
	}
	$self->Init();
	return $self;
}

sub Init
{
	my $self = shift;
	%{$self->{config}} = %{$self->{gotconfig}};
	$self->{comments} = []; #some comment in header
	$self->{titles} = []; #column titles
	$self->{title} = undef; #file title
	foreach my $k ('numregex','numberformat','comregex','quote')
	{
		$self->{config}->{$k} = $defaults{$k} unless defined $self->{config}->{$k};
	}
}

#LineCheck(\$line, $onlycheck) 
#$onlycheck: 0/undef: do full search for file/line titles and line end, etc.
#            1: only determine if data or not 
sub LineCheck #return 1 and set separator and line ending if data line and 0 otherwise
{
	my $self = shift;
	my $lr = shift;
	my $oc = shift;
	my $zahl = $self->{config}->{numregex};
	#the leading whitespace is a workaround for TISEAN
	#good or bad? It should not break any files that worked before...
	if( ${$lr} =~ /^\s*$/ and ${$lr} =~ /$nlend*($lend)$/ )
	{
		$self->{config}->{lineend} = $4 unless defined $self->{config}->{lineend};
		return 0;	
	}
	if(${$lr} =~ /^\s*($zahl)(($trenner)$nlend*|)($lend)$/)
	{
		$self->{config}->{lineend} = $4 unless not defined $4 or defined $self->{config}->{lineend};
		$self->{config}->{separator} = $3 unless (not defined $3 or defined $self->{config}->{separator} or $3 eq '');
		if($#{$self->{comments}} > -1 and $#{$self->{titles}} > -1){ pop(@{$self->{comments}}); }
		return 1;
	}
	elsif($oc){ return 0; }
	else
	{
		if(${$lr} =~ /^($self->{config}->{comregex})($lend)$/)
		{
			$self->{config}->{comment} = $1 unless defined $self->{config}->{comment};
			$self->{config}->{lineend} = $2 unless defined $self->{config}->{lineend};
		  return 0;
		}
		#first non-empty line is some kind of title or comment
		#first means: we didn't have content up to now
		if(${$lr} =~ /^($self->{config}->{comregex})($nlend+)($lend)$/)
		{
			if( defined $self->{title} ){ push(@{$self->{comments}},$2); }
			else{	$self->{title} = $2; }
			$self->{config}->{lineend} = $3 unless defined $self->{config}->{lineend};
			$self->{config}->{comment} = $1 unless defined $self->{config}->{comment};
		}
		#attention: I take " or ' just as quotes, do distinction!
		if(${$lr} =~ /^($self->{config}->{comregex})(["'])($nlend*["']($trenner)["']$nlend*)["']($lend)$/)
		{
			# "axis title"\t"axis title"\t"..."
			my @ax = split($2.$4.$2,$3);
			$self->{titles} = \@ax;
			$self->{config}->{lineend} = $5 unless defined $self->{config}->{lineend};
			$self->{config}->{comment} = $1 unless defined $self->{config}->{comment};
		}
		#either no quotes at all or maybe quotes but single item without separator
		elsif(${$lr} =~ /^($self->{config}->{comregex})["']?($nlend*)($lend)$/)
		{
			$self->{config}->{lineend} = $3 unless defined $self->{config}->{lineend};
			$self->{config}->{comment} = $1 unless defined $self->{config}->{comment};
			my $d = $2;
			$d =~ s/["']$//;
			my @ax = ();
			if($d =~ /($trenner)/)
			{
				@ax = split($1, $d);
			}
			else{ @ax = ($d); }
			$self->{titles} = \@ax;
		}
		return 0;
	}
}

sub LineData
{
	my $self = shift;
	my $lr = shift;
	my @ar = ();
	my $zahl = $self->{config}->{numregex};
	if($self->{config}->{strict})
	{
		#just split with defined or found separator
		@ar = split(defined $self->{config}->{separator} ? $self->{config}->{separator} : $defaults{separator}, ${$lr});
		#remove line end
		if($#ar > -1){ $ar[$#ar] =~ s/\r*\n*$//; }
	}
	else
	{
		my $l = ${$lr};
		if($self->{config}{black})
		{
			$l =~ s/^\s*//;
			# s/\s*$// deletes the line end -- no problem here
			$l =~ s/\s*$//;
		}
		if($l =~ /^($zahl)(.*)$/){ push(@ar, $1); $l = $2; }else{ return undef; }
		unless($self->{config}->{text})
		{
			while($l =~ /^($trenner)($zahl)(.*)$/)
			{
				push(@ar, $2);
				$l = $3;
			}
		}
		else
		{
			while($l =~ /^($trenner)($ntrenner)(.*)$/)
			{
				push(@ar, $2);
				$l = $3;
			}
		}
	}
	return \@ar;
}

sub DataLine
{
	my $self = shift;
	my $ar = shift;
	my $wr = shift;
	my $l = '';
	my $zahl = $self->{config}->{numregex};
	my $end = defined $self->{config}->{lineend} ? $self->{config}->{lineend} : $defaults{lineend};
	my $sep = defined $self->{config}->{outseparator} ? $self->{config}->{outseparator} : (defined $self->{config}->{separator} ? $self->{config}->{separator} : $defaults{separator});
	if(defined $self->{config}->{numberformat}->[0])
	{
		unless(defined $wr)
		{
			my $i = -1;
			foreach my $v (@{$ar})
			{
				$self->{config}->{numberformat}->[$i] = $self->{config}->{numberformat}->[0] unless defined $self->{config}->{numberformat}->[++$i];
				if($self->{config}->{numberformat}->[$i] ne '')
				{
					$l .= ($v ne '' and $v =~ /$zahl/ ? sprintf($self->{config}->{numberformat}->[$i], $v) : $v).$sep;
				}
				else{ $l .= $v.$sep; }
			}
		}
		else
		{
			foreach my $k (@{$wr})
			{
				$self->{config}->{numberformat}->[$k] = $self->{config}->{numberformat}->[0] unless defined $self->{config}->{numberformat}->[$k];
				if($self->{config}->{numberformat}->[$k] ne '')
				{
					$l .= ($ar->[$k] ne '' and $ar->[$k] =~ /$zahl/ ? sprintf($self->{config}->{numberformat}->[$k], $ar->[$k]) : $ar->[$k]).$sep if $k > -1;
				}
				else
				{
					$l .= $ar->[$k].$sep if $k > -1;
				}
			}
		}
		$l =~ s/$sep$/$end/;
	}
	else
	{
		unless(defined $wr)
		{
			$l = join($sep, @{$ar}).$end;
		}
		else
		{
			foreach my $k (@{$wr})
			{
				$l .= $ar->[$k].$sep if $k > -1;
			}
			$l =~ s/$sep$/$end/;
		}
	}
	return \$l;
}

sub TitleLine
{
	my $self = shift;
	my $ar = shift;

	my $end = defined $self->{config}->{lineend} ? $self->{config}->{lineend} : $defaults{lineend};
	my $sep = defined $self->{config}->{outseparator} ? $self->{config}->{outseparator} : (defined $self->{config}->{separator} ? $self->{config}->{separator} : $defaults{separator});
	my $com = defined $self->{config}->{comment} ? $self->{config}->{comment} : $defaults{comment};
	my $q = $self->{config}{quote} ? '"' : '';
	my $l = $com.$q;
	#print STDERR "titles: @{$self->{titles}}\n";
	#print STDERR "titles for @{$ar}\n" if defined $ar;
	unless(defined $ar){ $l = $com.$q.join($q.$sep.$q, @{$self->{titles}}).$q.$end; }
	else
	{
		foreach my $k (@{$ar})
		{
			$l .= $self->{titles}->[$k].$q.$sep.$q if $k > -1;
		}
		$l =~ s/$q$//;
		$l =~ s/$sep$/$end/;
	}
	return \$l;
	
}

sub CommentLine
{
	my $self = shift;
	my $line = shift;
	my $cline = (defined $self->{config}->{comment} ? $self->{config}->{comment} : $defaults{comment}).${$line}.(defined $self->{config}->{lineend} ? $self->{config}->{lineend} : $defaults{lineend});
	return \$cline;
}

sub Chomp
{
	my $self = shift;
	if(defined (my $string = shift))
	{
		${$string} =~ s/$lend$//;
	}
}

sub MakeNaked
{
	my $self = shift;
	if(defined (my $string = shift))
	{
		${$string} =~ s/$lend$//;
		${$string} =~ s/^$self->{config}->{comregex}//;
	}
}


1;

__END__

=head1 NAME

TextData::Lines - Basic parsing of text data lines

=head1 SYNOPSIS

	use TextData::Lines;
	my $c = new TextData::Lines;
	my $line = "6e-6 3.4e7 123 0\n";
	my $data = $c->LineData(\$line);

	$data->[3] *= $data->[0]*$data->[1];
	
	print $c->DataLine($data);

=head1 DESCRIPTION

This module (class) contains the basic parsing structure for TextData. It is intended for use with numerical data sets in text files as commonly produced by data aquisition software - there often called "ASCII file". It's not about arbitary tabular data - the main intention are rows of numbers after some header that ideally contains information about what kind of data is in there. Simple text fields are supported, but beware of any separator-looking characters in there!
The provided function for checking if there is data or header relies on the first data item being a number.
Additionall, the following general header layout is assumed:

	#file title
	#comments
	#comments
	#...
	#row titles with proper separators with or without quotes

Comments are optional, only a single non-empty header line results in both file and row titles to be deduced from the very same line.

=head1 MEMBERS

Some members of the hash repesenting an TextData::Lines object:

=head2 Methods (Functions)

=over * 4

=item * LineCheck(\$line[, $onlycheck]) -> 0 or 1

checks if $line appears to have some data (starts with a number followed by a separator or nothing) and scans for various stuff items as line ending, separator, file and data row titles... the scanning is disabled if $onlycheck is true; returns 1 if $line supposedly is containing data

=item * LineData(\$line) -> \@data

input: extract the @data of $line

=item * DataLine(\@data) -> \$line

output: have @data, produce $line

=item * CommentLine(\$comment) -> \$line

output: have bare comment, produce full $line

=item * TitleLine(\@colnumbers) -> \$line

output: form a proper line with the titles matching the specified columns (starting at 0 - plain array index!) or a line for all columns if @colnumbers is unspecified 

=item * Chomp(\$line)

Just what you would expect from a chomp()... making sure that really any kind of line ending (that matches the internal regex) is chopped off; modifies the input directly

=item * MakeNaked(\$line)

Chomp + remove comment characters

=back

=head2 Data

=over * 4

=item * title

file title

=item * comments

array ref of comment lines (w/o comment character(s) and line ending)

=item * titles

array ref of data row titles

=back

=head1 CONFIGURATION

You can provide a hash reference with configuration data to the constructor to tweak some aspects of the created object. They have (mostly) sensible defaults and/or are normally deduced from the input data (p.ex. line end, separator) if possible. 
You can always check the configuration by looking at the $c->{config} hash reference.
The different parameters fall roughly into two categories:

=head2 Parsing

These influence how the data is parsed. A considerable amount of work went into the regexes... to change them properly you better should know and understand the source code of this module.

=over * 4

=item * numregex

the regular expression a number has to match (I really hope that the default is reasonable!)

=item * comregex

regex for beginning of comment line (additional to not starting with a number)

=item * strict [0/1]

just split at every separator occurence to get the data array (otherwise there is some more fuzzy logic with treating multiple separation characters as one separation)

=item * text [0/1]

allow text in data field in non-strict mode (no effect in strict mode)

=back

=head2 Output

=over * 4

=item * separator

The separator to use for data fields/columns; when strict is set it is also used for the actual data parsing.	
	
=item * lineend

line ending to use

=item * comment

character(s) to put in front of comment lines

=item * numberformat
	
listref with formats for numbers as sprintf format strings (p.ex. %02d), one list entry is for one data row

=back 

=head1 LICENSE

This package is open source software under the terms of the Artistic License. Free use respecting the author;-)

=head1 AUTHOR

Thomas Orgis <thomas@orgis.org>

=head1 WEB

See	http://thomas.orgis.org/scripterei for updates and lots of other (un)related stuff.

